/* scatter.c - Compute particle scattering properties using a range of methods
 * 
 * Author: Robin Hogan <r.j.hogan@ecmwf.int>
 * Copyright (C) 2009-2015  University of Reading
 * Copyright (C) 2016  European Centre for Medium Range Weather Forecasts 
 * Language: C99
 *
 * This program reads an ascii configuration file from the
 * command-line, optionally modified with param=value pairs, and
 * outputs a NetCDF file containing the computed scattering and
 * terminal fall-speed properties. It can use Mie theory, the T-matrix
 * method (for horizontally aligned particles) or the Self-Similar
 * Rayleigh Gans Approximation. It can also interpolate from an
 * existing look-up table of scattering properties.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <math.h>
#include <netcdf.h>

/* This is generated by autoconf - currently nothing from it is
   used */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "pi.h"
#include "refractive_index.h"
#include "bhmie.h"
#include "tmatrix.h"
#include "readconfig.h"
#include "nctools.h"
#include "maxwell_garnett.h"
#include "beard_droplet_fall_speed.h"
#include "heymsfield_westbrook_ice_fall_speed.h"
#include "legendre_decomposition.h"
#include "self_similar_rayleigh_gans.h"
#include "ssrga.h"

#define MAX_STRING_LENGTH 1000

typedef double real;
#define NC_PUT_VAR_REAL nc_put_var_double
#define NC_PUT_VAR1_REAL nc_put_var1_double
#define NC_PUT_VARA_REAL nc_put_vara_double
#define NC_GET_VAR_REAL nc_get_var_double

#define SUCCESS 0
#define FAILURE 1

/* Particle type, specified by "medium" */
#define UNSPECIFIED -1
#define UNKNOWN 0
#define WATER 1
#define ICE 2

/* Computational method, specified by "engine" */
#define BHMIE 1
#define AREA_SCALED_BHMIE 2
#define TMATRIX 3
#define SELF_SIMILAR_RAYLEIGH_GANS_ORIG 4
#define SELF_SIMILAR_RAYLEIGH_GANS 5
#define AREA_LOOK_UP_TABLE 6

/* Number of angles between 0 and 90 to compute phase function */
#define NANG_TO_90 901
static const int nang_to_90 = NANG_TO_90;
static const int nang = 2*NANG_TO_90 - 1;

/* Reference temperature and air-density at which to perform droplet
   fall-speed calculations; if these are changed then the comment to
   the fall_speed coordinate variable should be changed too. */
static const real TEMPERATURE_DROP_FALLSPEED_REF = 283.15;
static const real AIR_DENSITY_DROP_FALLSPEED_REF = 1.0;
static const real AIR_DENSITY_ICE_FALLSPEED_REF = 1.0;

/* Check the return value from a call to a NetCDF function and quit
   semi-elegantly if an error occurred. */
int ncstatus;
#define NC_CHECK(a) if ((ncstatus = (a)) != NC_NOERR) {		     \
    fprintf(stderr, "NetCDF error on line %d of %s: %s\n", __LINE__, \
	    __FILE__, nc_strerror(ncstatus));			     \
    exit(1);							     \
  }

/* Return an allocated vector of real numbers for variable <varname>
   (e.g. "temperature") from the configuration data. If <varname> is
   not present then a vector of length <varname>_values from
   <varname>_start to <varname>_end will be constructed, and if
   <varname>_scale is present and starts with "log" then the values
   will be logarithmically spaced.  */
static
int
get_vector(rc_data* config, char* varname,
	   real** var, int* n)
{
  /* Get wavelength information */
  if (!(*var = rc_get_real_vector(config, varname, n))) {
    /* No variable by the name varname: find range information
       instead */
    char buffer[MAX_STRING_LENGTH];
    char* scale = NULL;
    real start, end;
    int len = strlen(varname);
    int log_scale = 0;

    if (len > MAX_STRING_LENGTH-20) {
      fprintf(stderr, "Error: requested variable name \"%s\" is too long\n", varname);
      exit(1);
    }
    strcpy(buffer, varname);
    /* Find the number of values */
    strcpy(buffer+len, "_values");
    if (!rc_assign_int(config, buffer, n)) {
      return FAILURE;
    }
    /* Allocate memory */
    *var = malloc(sizeof(real)*(*n));
    if (!*var) { 
      fprintf(stderr, "Error: unable to allocate memory for %d reals\n", *n);
      exit(1);
    }
    strcpy(buffer+len, "_start");
    if (!rc_assign_real(config, buffer, &start)) {
      /* fprintf(stderr, "Error: \"%s\" not found\n", buffer);
	 exit(1); */
      return FAILURE;
    }
    strcpy(buffer+len, "_end");
    if (!rc_assign_real(config, buffer, &end)) {
      /* fprintf(stderr, "Error: \"%s\" not found\n", buffer);
	 exit(1); */
      return FAILURE;
    }
    strcpy(buffer+len, "_scale");
    if ((scale = rc_get_string(config, buffer))) {
      if (strncasecmp(scale, "log", 3) == 0) {
	log_scale = 1;
      }
      rc_free(scale);
    }
    if (log_scale) {
      int num = *n;
      int i;
      real log_start = log(start);
      real factor = (log(end)-log_start)/(num-1);
      for (i = 0; i < num; i++) {
	(*var)[i] = exp(log(start) + i*factor);
      }
    }
    else {
      int num = *n;
      int i;
      real factor = (end-start)/(num-1);
      for (i = 0; i < num; i++) {
	(*var)[i] = start+i*factor;
      }
    }
  }
  return SUCCESS;
}

/* Structure for holding a look-up table (e.g. created using ray
   tracing) held in an external netcdf file */
typedef struct {
  real* area;
  real* wavelength;
  real* ext;
  real* scat;
  real* bscat;
  real* g;
  size_t nsize, nwavelength;
} Lut;

/* Load a look-up table from a netcdf file */
static
void
load_lut(const char* filename, Lut* lut) {
  int ncid, sizeid, wavelengthid, varid;
  fprintf(stdout, "Reading look-up table file %s", filename);
  NC_CHECK(nc_open(filename, 0, &ncid));
  NC_CHECK(nc_inq_dimid(ncid, "size", &sizeid));
  NC_CHECK(nc_inq_dimid(ncid, "wavelength", &wavelengthid));
  NC_CHECK(nc_inq_dimlen(ncid, sizeid, &(lut->nsize)));
  NC_CHECK(nc_inq_dimlen(ncid, wavelengthid, &(lut->nwavelength)));
  fprintf(stdout, " ...contains %zu wavelengths and %zu sizes\n",
	  lut->nwavelength, lut->nsize);
  lut->area  = malloc(lut->nsize * sizeof(real));
  lut->wavelength = malloc(lut->nwavelength * sizeof(real));
  lut->ext   = malloc(lut->nsize * lut->nwavelength * sizeof(real));
  lut->scat  = malloc(lut->nsize * lut->nwavelength * sizeof(real));
  lut->bscat = malloc(lut->nsize * lut->nwavelength * sizeof(real));
  lut->g     = malloc(lut->nsize * lut->nwavelength * sizeof(real));
  if (!lut->area || !lut->wavelength || !lut->ext
      || !lut->scat || !lut->bscat || !lut->g) {
    fprintf(stderr, "Error allocating memory to read %s\n", filename);
    exit(1);
  }
  NC_CHECK(nc_inq_varid(ncid, "area", &varid));
  NC_CHECK(NC_GET_VAR_REAL(ncid, varid, lut->area));
  NC_CHECK(nc_inq_varid(ncid, "wavelength", &varid));
  NC_CHECK(NC_GET_VAR_REAL(ncid, varid, lut->wavelength));
  NC_CHECK(nc_inq_varid(ncid, "ext", &varid));
  NC_CHECK(NC_GET_VAR_REAL(ncid, varid, lut->ext));
  NC_CHECK(nc_inq_varid(ncid, "scat", &varid));
  NC_CHECK(NC_GET_VAR_REAL(ncid, varid, lut->scat));
  NC_CHECK(nc_inq_varid(ncid, "bscat", &varid));
  NC_CHECK(NC_GET_VAR_REAL(ncid, varid, lut->bscat));
  NC_CHECK(nc_inq_varid(ncid, "g", &varid));
  NC_CHECK(NC_GET_VAR_REAL(ncid, varid, lut->g));
  NC_CHECK(nc_close(ncid));
}

/* Look up the scattering properties for a given area and wavelength */
static
void
look_up(const Lut* lut, real wavelength, real area,
	float* ext, float* scat, float* bscat, float* g)
{
  int iwavelength = 0;
  int isize = 0;

  if (wavelength < lut->wavelength[0]
      || wavelength > lut->wavelength[lut->nwavelength-1]) {
    fprintf(stderr, "Error: look-up %g m outside wavelength range %g-%g m\n",
	    wavelength, lut->wavelength[0], lut->wavelength[lut->nwavelength-1]);
    exit(1);
  }

  while (iwavelength < lut->nwavelength-2
	 && wavelength > lut->wavelength[iwavelength+1]) {
    iwavelength++;
  }
  while (isize < lut->nsize-2
	 && area > lut->area[isize+1]) {
    isize++;
  }

  int i00 = iwavelength*lut->nsize + isize;
  int i01 = iwavelength*lut->nsize + isize+1;
  int i10 = (iwavelength+1)*lut->nsize + isize;
  int i11 = (iwavelength+1)*lut->nsize + isize+1;

  real wt_wavelength = (lut->wavelength[iwavelength+1]-wavelength)
    / (lut->wavelength[iwavelength+1]-lut->wavelength[iwavelength]);

  if (area < lut->area[0]) {
    /* Size too small: interpolate to zero; linear interpolation in
       wavelength */
    real frac = area / lut->area[0];
    *ext   = frac * (lut->ext[i00]  *wt_wavelength + lut->ext[i10]  *(1.0-wt_wavelength));
    *scat  = frac * (lut->scat[i00] *wt_wavelength + lut->scat[i10] *(1.0-wt_wavelength));
    *bscat = frac * (lut->bscat[i00]*wt_wavelength + lut->bscat[i10]*(1.0-wt_wavelength));
    *g     =         lut->g[i00]    *wt_wavelength + lut->g[i10]    *(1.0-wt_wavelength);
  }
  else if (area > lut->area[lut->nsize-1]) {
    /* Size too large: assume proportionality to area; linear
       interpolation in wavelength */
    real frac = area / lut->area[lut->nsize-1];
    *ext   = frac * (lut->ext[i01]  *wt_wavelength + lut->ext[i11]  *(1.0-wt_wavelength));
    *scat  = frac * (lut->scat[i01] *wt_wavelength + lut->scat[i11] *(1.0-wt_wavelength));
    *bscat = frac * (lut->bscat[i01]*wt_wavelength + lut->bscat[i11]*(1.0-wt_wavelength));
    *g     =         lut->g[i01]    *wt_wavelength + lut->g[i11]    *(1.0-wt_wavelength);  
  }
  else {
    /* Bilinear interpolation */
    real wt_size = (lut->area[isize+1]-area) / (lut->area[isize+1]-lut->area[isize]);
    *ext = wt_wavelength * (lut->ext[i00]  * wt_size
			    +lut->ext[i01] * (1.0 - wt_size))
      +(1.0-wt_wavelength)*(lut->ext[i10]  * wt_size
			    +lut->ext[i11] * (1.0 - wt_size));
    *scat = wt_wavelength * (lut->scat[i00]  * wt_size
			    +lut->scat[i01] * (1.0 - wt_size))
      +(1.0-wt_wavelength)*(lut->scat[i10]  * wt_size
			    +lut->scat[i11] * (1.0 - wt_size));
    *bscat = wt_wavelength * (lut->bscat[i00]  * wt_size
			    +lut->bscat[i01] * (1.0 - wt_size))
      +(1.0-wt_wavelength)*(lut->bscat[i10]  * wt_size
			    +lut->bscat[i11] * (1.0 - wt_size));
    *g = wt_wavelength * (lut->g[i00]  * wt_size
			    +lut->g[i01] * (1.0 - wt_size))
      +(1.0-wt_wavelength)*(lut->g[i10]  * wt_size
			    +lut->g[i11] * (1.0 - wt_size));
  }
}

/* The main program */
int
main(int argc, const char** argv)
{
  real pi = 4.0*atan(1.0);
  rc_data* config = NULL;
  char* medium_str = NULL;
  char* engine_str = NULL;

  int n_wavelength = 0;
  real* wavelength = NULL;
  int i_wavelength = -1;

  int n_size = 0;
  real* size = NULL;
  char* size_variable = NULL;
  real size_scale = 1.0;
  int i_size = -1;

  int n_temperature = 0;
  real* temperature = NULL;
  int i_temperature = -1;

  real default_fraction = 1.0;
  int n_fraction = 0;
  real* fraction = 0;
  real max_density = 0.0;
  int i_fraction = -1;

  /* Total number of particles to compute */
  int n_total = 0; 
  /* Number of current particle */
  int i_particle = 0;
  int i_last_dot = 0;

  char* output_filename = "scat.nc";

  char* lut_filename;

  real* refind_r = NULL;
  real* refind_i = NULL;
  int n_refind_r = 0;
  int n_refind_i = 0;

  float phase_function[nang];

  int engine = UNKNOWN;
  int medium = UNSPECIFIED;

  real axial_ratio = 1.0;
  real axial_ratio_monomer = 1.0;

  /* Geometric cross sectional area = prefactor * diameter ^ exponent;
     the default assumes spheres with no gaps */
  real area_diameter_prefactor = 0.25*pi;
  real area_diameter_exponent = 2.0;
  real* area = NULL;

  int ncid;
  int ext_id, scat_id, bscat_id, g_id;
  int wavelength_id, temperature_id, fraction_id;
  int refind_r_id, refind_i_id;
  int size_id, area_id, mass_id, fall_speed_id, dmax_id;
  int phase_function_angle_id, phase_function_id;
  int legendre_coefficient_id;
  int wavelength_dimid = 0, temperature_dimid = 0;
  int fraction_dimid = 0, size_dimid = 0;
  int phase_function_angle_dimid = 0;
  int legendre_coefficient_dimid = 0;
  int dimids[6];
  int ndims = 0;
  size_t index[6];

  int i, j, k, l;

  int show_dots = 0;
  int calc_fall_speed = 0;
  int compute_phase_function = 0;
  int n_legendre_coefficients = 0;
  int n_phase_function_angles = 0;

  /* Self-similar Rayleigh Gans parameters */
  double ssrga_alpha_eff = 1.0;
  double ssrga_kappa = 0.0;
  double ssrga_beta = 0.5;
  double ssrga_gamma = 2.0;
  double ssrga_zeta1 = 1.0;

  Lut lut;

  /* PARSE CONFIG INFORMATION */

  /* Find the first config file on the command line. */
  int ifile = rc_get_file(argc, argv);

  if (!ifile) {
    /* No file given - assume command-line arguments contain all the
       information. */
    config = rc_read(NULL, stderr);
    if (!config->param) {
      fprintf(stderr, "Usage: %s [param1=value1] [param2=value2...] file.cfg\n",
	      argv[0]);
      return 1;	      
    }
  }
  else {
    /* Read configuration information from the file. */
    config = rc_read(argv[ifile], stderr);
  }
  if (!config) {
    fprintf(stderr, "Error initializing configuration information\n");
    return 1;
  }
  
  /* Supplement configuration information with command-line
     arguments. */
  rc_register_args(config, argc, argv);

  /* Get temperature information */
  if (get_vector(config, "temperature",
		 &temperature, &n_temperature) == FAILURE) {
    fprintf(stderr, "Error reading \"temperature\" as a vector of reals\n");
    return 1;
  }

  /* Get medium information */
  if (!(medium_str = rc_get_string(config, "medium"))) {
    fprintf(stderr, "Error reading \"medium\" as a string\n");
    return 1;
  }
  else if (strcasecmp(medium_str, "water") == 0
	   || strncasecmp(medium_str, "liquid",6) == 0) {
    medium = WATER;
    max_density = 1000.0;
    calc_fall_speed = 1;
  }
  else if (strcasecmp(medium_str, "ice") == 0) {
    medium = ICE;
    max_density = 917.0;
    calc_fall_speed = 1;
  }
  else {
    /*
    fprintf(stderr, "Error: medium \"%s\" not supported\n",
	    medium_str);
    */
    medium = UNKNOWN;
    if (!rc_assign_real(config, "density", &max_density)
	|| !(n_refind_r = rc_assign_real_vector(config, "refind_r",
						&refind_r, 0))
	|| !(n_refind_i = rc_assign_real_vector(config, "refind_i",
						&refind_i, 0))) {
      fprintf(stderr, "Medium \"%s\" is unknown, so must be accompanied by variables \"density\", \"refind_r\" and \"refind_i\"\n",
	      medium_str);
      return 1;
    }
    if (n_refind_i != n_refind_r) {
      fprintf(stderr, "Sizes of real and imaginary parts of refractive index are not equal\n");
      return 1;
    }
    if (n_refind_r != n_temperature && n_refind_r != 1) {
      fprintf(stderr, "Refractive index is not the same size as temperature field\n");
      return 1;
    }
  }

  /* Get engine information */
  if (!(engine_str = rc_get_string(config, "engine"))) {
    fprintf(stderr, "Error reading \"engine\" as a string\n");
    return 1;
  }
  else if (strcasecmp(engine_str, "bhmie") == 0) {
    engine = BHMIE;
  }
  else if (strcasecmp(engine_str, "area_scaled_bhmie") == 0) {
    engine = AREA_SCALED_BHMIE;
  }
  else if (strcasecmp(engine_str, "tmatrix") == 0) {
    engine = TMATRIX;
    rc_assign_real(config, "axial_ratio", &axial_ratio);
  }
  else if (strcasecmp(engine_str, "self_similar_rayleigh_gans_orig") == 0) {
    engine = SELF_SIMILAR_RAYLEIGH_GANS_ORIG;
    rc_assign_real(config, "axial_ratio", &axial_ratio);
    rc_assign_real(config, "ssrga_kappa", &ssrga_kappa);
    rc_assign_real(config, "ssrga_beta", &ssrga_beta);
    rc_assign_real(config, "ssrga_gamma", &ssrga_gamma);
  }
  else if (strcasecmp(engine_str, "self_similar_rayleigh_gans") == 0) {
    engine = SELF_SIMILAR_RAYLEIGH_GANS;
    rc_assign_real(config, "axial_ratio", &axial_ratio);
    rc_assign_real(config, "axial_ratio_monomer", &axial_ratio_monomer);
    rc_assign_real(config, "ssrga_alpha_eff", &ssrga_alpha_eff);
    rc_assign_real(config, "ssrga_kappa", &ssrga_kappa);
    rc_assign_real(config, "ssrga_beta", &ssrga_beta);
    rc_assign_real(config, "ssrga_gamma", &ssrga_gamma);
    rc_assign_real(config, "ssrga_zeta1", &ssrga_zeta1);
  }
  else if (strcasecmp(engine_str, "area_look_up_table") == 0) {
    engine = AREA_LOOK_UP_TABLE;
    if (!(lut_filename = rc_get_string(config, "lut_filename"))) {
      fprintf(stderr, "Error reading \"lut_filename\" as a string\n");
      return 1;
    }
    load_lut(lut_filename, &lut);    
  }
  else {
    fprintf(stderr, "Error: scattering engine \"%s\" not supported\n",
	    engine_str);
    return 1;
  }

  /* Get information about phase-function output required */
  compute_phase_function = rc_get_boolean(config, "output_phase_function");
  if (compute_phase_function) {
    n_phase_function_angles = nang;
  }
  if (rc_assign_int(config, "legendre_polynomial_coefficients", 
		    &n_legendre_coefficients)) {
    compute_phase_function = 1;
  }

  if (engine == TMATRIX && compute_phase_function) {
    fprintf(stderr, "Cannot currently compute phase function from T-matrix\n");
    return 1;
  }

  /* Get wavelength information */
  if (get_vector(config, "wavelength", &wavelength, &n_wavelength)
      == FAILURE) {
    /* No wavelength present; try wavenumber */
    if (get_vector(config, "wavenumber_cm_1", &wavelength, &n_wavelength)
	== SUCCESS) {
      for (i = 0; i < n_wavelength; ++i) {
	/* Convert wavenumber to wavelength */
	wavelength[i] = 0.01 / wavelength[i];
      }
    }
    else {
      fprintf(stderr, "Error reading \"wavelength\" or \"wavenumber_cm_1\" as a vector of reals\n");
      return 1;
    }
  }

  /* Get size information */
  if (get_vector(config, "size", &size, &n_size) == FAILURE) {
    fprintf(stderr, "Error reading \"size\" as a vector of reals\n");
    return 1;
  }

  if (!(size_variable = rc_get_string(config, "size_variable"))) {
    fprintf(stderr, "Error reading \"size_variable\" as a string\n");
    return 1;
  }
  else if (strcasecmp(size_variable, "radius") == 0) {
    size_scale = 2.0;
    /* Scale it to make it diameter/mean_dimension */
    int i;
    for (i = 0; i < n_size; i++) {
      size[i] *= size_scale;
    }
    size_scale = 1.0;
  }
  else if (strcasecmp(size_variable, "diameter") == 0
	   || strcasecmp(size_variable, "mean_dimension") == 0) {
    size_scale = 1.0;
  }
  else {
    fprintf(stderr, "Error: size variable \"%s\" not understood\n",
	    size_variable);
    return 1;
  }

  if (axial_ratio > 1.0) {
    fprintf(stderr, "Error: axial_ratio must be less than or equal to one\n");
    exit(1);
  }

  /* Maximum dimension, useful for ice */
  real Dmax[n_size];
  {
    real Dmax_scale = pow(axial_ratio, -1.0/3.0);
    for (i = 0; i < n_size; i++) {
      Dmax[i] = size[i] * Dmax_scale;
    }
  }

  /* Get fraction information */
  if (get_vector(config, "fraction", &fraction, &n_fraction) == FAILURE) {
    fraction = &default_fraction;
  }

  /* Get output filename */
  rc_assign_string(config, "output", &output_filename);

  /* Get coefficients of area-diameter relationship */
  rc_assign_real(config, "area_diameter_prefactor", &area_diameter_prefactor);
  rc_assign_real(config, "area_diameter_exponent", &area_diameter_exponent);

  /* Assign geometric cross-sectional area */
  area = malloc(n_size*sizeof(real));
  if (!area) {
    fprintf(stderr, "Error: unable to allocate memory for %d reals to store areas\n", n_size);
    exit(1);
  }
  for (i = 0; i < n_size; i++) {
    /* Size can be assumed to be diameter in metres */
    real new_area = area_diameter_prefactor*pow(size[i], area_diameter_exponent);
    area[i] = 0.25*pi*size[i]*size[i]; /* Diameter of a sphere */
    if (new_area < 0.05*area[i]) {
      /* Minimum area is 5% of the area a sphere of the same diameter
	 would have */
      area[i] *= 0.05;
    }
    else if (new_area < area[i]) {
      /* Only use the power law for area if it is less than that of a
	 sphere */
      area[i] = new_area;
    } 
  }
  
  /* Do we show dots for each size calculated? */
  show_dots = rc_get_boolean(config, "show_dots");

  /* Report what the program is doing */
  fprintf(stdout, "Scattering calculations for %s: ", medium_str);
  if (n_wavelength > 1) {
    fprintf(stdout, "%d wavelengths", n_wavelength);
  }
  else {
    fprintf(stdout, "wavelength %g m", wavelength[0]);
  }
  if (n_temperature > 1) {
    fprintf(stdout, ", %d temperatures", n_temperature);
  }
  else {
    fprintf(stdout, ", temperature %g K", temperature[0]);
  }
  if (n_fraction > 1) {
    fprintf(stdout, ", %d fractions", n_fraction);
  }
  if (n_size > 1) {
    fprintf(stdout, ", %d sizes\n", n_size);
  }
  else {
    fprintf(stdout, ", size %g m\n", size[0]);
  }


  /* INITIALIZE NETCDF FILE */

  /* Open file */
  NC_CHECK(nc_create(output_filename, NC_CLOBBER, &ncid));

  /* Dimensions */
  if (n_wavelength > 1) {
    NC_CHECK(nc_def_dim(ncid, "wavelength",
			n_wavelength, &wavelength_dimid));
    dimids[(i_wavelength = ndims++)] = wavelength_dimid;
  }
  if (n_temperature > 1) {
    NC_CHECK(nc_def_dim(ncid, "temperature",
			n_temperature, &temperature_dimid));
    dimids[(i_temperature = ndims++)] = temperature_dimid;
  }
  if (n_fraction > 1) {
    NC_CHECK(nc_def_dim(ncid, "fraction",
			n_fraction, &fraction_dimid));
    dimids[(i_fraction = ndims++)] = fraction_dimid;
  }
  NC_CHECK(nc_def_dim(ncid, "size",
		      n_size, &size_dimid));
  dimids[(i_size = ndims++)] = size_dimid;

  if (n_phase_function_angles > 0) {
    NC_CHECK(nc_def_dim(ncid, "phase_function_angle",
			n_phase_function_angles, 
			&phase_function_angle_dimid));
  }
  if (n_legendre_coefficients > 0) {
    NC_CHECK(nc_def_dim(ncid, "legendre_coefficient",
			n_legendre_coefficients, 
			&legendre_coefficient_dimid));
  }

  /* Coordinate variables */
  NC_CHECK(nc_def_var(ncid, "wavelength", NC_FLOAT, n_wavelength>1,
		      &wavelength_dimid, &wavelength_id));
  NC_CHECK(nct_add_standard_attributes(ncid, wavelength_id, "m",
				       "Wavelength", NULL, NULL));
  NC_CHECK(nc_def_var(ncid, "temperature", NC_FLOAT,
		      n_temperature>1, &temperature_dimid,
		      &temperature_id));
  NC_CHECK(nct_add_standard_attributes(ncid, temperature_id, "K",
				       "Temperature", NULL, NULL));

  if (n_fraction > 0) {
    NC_CHECK(nc_def_var(ncid, "fraction", NC_FLOAT,n_fraction>1,
			&fraction_dimid, &fraction_id));
    NC_CHECK(nct_add_standard_attributes(ncid, fraction_id, "1",
	 "Volume fraction of the particle that is ",NULL,NULL));
    NC_CHECK(nct_append_attribute(ncid, fraction_id, "long_name",
				  medium_str, ""));
    NC_CHECK(nct_append_attribute(ncid, fraction_id, "long_name",
				  ", the rest being air", ""));
  }

  if (n_phase_function_angles > 0) {
    NC_CHECK(nc_def_var(ncid, "phase_function_angle", NC_FLOAT,1,
			&phase_function_angle_dimid,
			&phase_function_angle_id));
    NC_CHECK(nct_add_standard_attributes(ncid, phase_function_angle_id, 
					 "degrees","Scattering angle",
					 NULL, NULL));
  }

  /* Coordinate-like variables */
  NC_CHECK(nc_def_var(ncid, "mean_dimension", NC_FLOAT,
		      1, &size_dimid, &size_id));
  NC_CHECK(nc_def_var(ncid, "maximum_dimension", NC_FLOAT,
		      1, &size_dimid, &dmax_id));
  if (axial_ratio < 1.0) {
    NC_CHECK(nct_add_standard_attributes(ncid, size_id, "m",
					 "Mean dimension", NULL,
       "This is the equivalent-volume diameter of an enclosing spheroid, so for an oblate\n"
       "spheroid with maximum dimension Dmax and minimum dimension Dmin, it is defined as\n"
       "cube_root(Dmax*Dmax*Dmin)."));
    NC_CHECK(nct_add_standard_attributes(ncid, dmax_id, "m",
					 "Maximum dimension", NULL,
					 NULL));
  }
  else {
    NC_CHECK(nct_add_standard_attributes(ncid, size_id, "m",
					 "Mean dimension", NULL,
       "For spheres or quasi-spherical particles this is equal to diameter."));
    NC_CHECK(nct_add_standard_attributes(ncid, dmax_id, "m",
					 "Maximum dimension", NULL,
       "For spheres or quasi-spherical particles this is equal to diameter."));
  }

  NC_CHECK(nc_def_var(ncid, "area", NC_FLOAT,
		      1, &size_dimid, &area_id));
  NC_CHECK(nct_add_standard_attributes(ncid, area_id, "m2",
				       "Cross-sectional area", NULL,
       "The geometric projected area of the particle when viewed from zenith or nadir."));
  NC_CHECK(nc_def_var(ncid, "mass", NC_FLOAT,
		      1 + (n_fraction>1),
		      &dimids[ndims-1-(n_fraction>1)], &mass_id));
  NC_CHECK(nct_add_standard_attributes(ncid, mass_id, "kg",
				       "Mass", NULL, NULL));

  if (calc_fall_speed) {
    /* This will need to be changed when ice fall speed is available */
    NC_CHECK(nc_def_var(ncid, "fall_speed", NC_FLOAT,
			1 + (n_fraction>1),
			&dimids[ndims-1-(n_fraction>1)], &fall_speed_id));
    if (medium == WATER) {
      NC_CHECK(nct_add_standard_attributes(ncid, fall_speed_id, "m s-1",
					   "Terminal fall speed", NULL, 
	 "The terminal fall speed at a reference temperature of 283.15 K and reference air density\n"
         "of 1.0 kg m-3, calculated using the equations of Beard (J. Atmos. Sci. 1976 p851-)."));
    }
    else if (medium == ICE) {
      NC_CHECK(nct_add_standard_attributes(ncid, fall_speed_id, "m s-1",
					   "Terminal fall speed", NULL, 
	 "The terminal fall speed at a reference air density of 1.0 kg m-3, calculated using\n"
         "the equations of Heymsfield and Westbrook (J. Atmos. Sci. 2010 p2469-)."));
    }
  }

  NC_CHECK(nc_def_var(ncid, "n_r", NC_FLOAT, 
		      (n_wavelength>1) + (n_temperature>1) + (n_fraction>1),
		      dimids, &refind_r_id));
  NC_CHECK(nct_add_standard_attributes(ncid, refind_r_id, "1",
       "Real part of refractive index", NULL, NULL));
  NC_CHECK(nc_def_var(ncid, "n_i", NC_FLOAT, 
		      (n_wavelength>1) + (n_temperature>1) + (n_fraction>1),
		      dimids, &refind_i_id));
  NC_CHECK(nct_add_standard_attributes(ncid, refind_i_id, "1",
       "Imaginary part of refractive index", NULL, 
       "Note that we adopt the convention of positive imaginary part."));

  /* Scattering property variables */
  NC_CHECK(nc_def_var(ncid, "ext", NC_FLOAT, ndims,
		      dimids, &ext_id));
  NC_CHECK(nct_add_standard_attributes(ncid, ext_id, "m2",
				       "Extinction cross-section",
				       NULL, NULL));
  NC_CHECK(nc_def_var(ncid, "scat", NC_FLOAT, ndims,
		      dimids, &scat_id));
  NC_CHECK(nct_add_standard_attributes(ncid, scat_id, "m2",
				       "Scattering cross-section",
				       NULL, NULL));
  NC_CHECK(nc_def_var(ncid, "bscat", NC_FLOAT, ndims,
		      dimids, &bscat_id));
  NC_CHECK(nct_add_standard_attributes(ncid, bscat_id, "m2 sr-1",
				       "Backscatter cross-section",
				       NULL,
       "This is the backscatter cross-section for zenith or nadir incidence. Note that this\n"
       "variable is defined such that if the particle with backscatter cross-section B is\n"
       "illuminated by radiation from a particular direction with intensity F W/m2, then the\n"
       "power scattered into the backwards direction per steradian will be F*B W/sterad.\n"
       "Multiply this variable by 4pi to obtain the radar cross-section."));
  NC_CHECK(nc_def_var(ncid, "g", NC_FLOAT, ndims,
		      dimids, &g_id));
  NC_CHECK(nct_add_standard_attributes(ncid, g_id, "1",
				       "Asymmetry factor", NULL, 
	       "The mean of the cosine of the scattering angle."));

  if (n_phase_function_angles > 0) {
    dimids[ndims] = phase_function_angle_dimid;
    NC_CHECK(nc_def_var(ncid, "phase_function", NC_FLOAT, ndims+1,
			dimids, &phase_function_id));
    NC_CHECK(nct_add_standard_attributes(ncid, phase_function_id, "1",
					 "Scattering phase function", NULL,
	 "This is normalized such that the integral from 0 to pi of the phase function\n"
	 "multiplied by the sine of the scattering angle is one."));
  }
  if (n_legendre_coefficients > 0) {
    dimids[ndims] = legendre_coefficient_dimid;
    NC_CHECK(nc_def_var(ncid, "legendre_coefficients", NC_FLOAT, ndims+1,
			dimids, &legendre_coefficient_id));
    NC_CHECK(nct_add_standard_attributes(ncid, legendre_coefficient_id, "1",
					 "Legendre coefficients", NULL, NULL));
  }

  NC_CHECK(nct_add_string_attribute(ncid, NC_GLOBAL, "title",
				    "Scattering properties of"));
  switch (medium) {
  case ICE:
    NC_CHECK(nct_append_attribute(ncid, NC_GLOBAL, "title", "ice particles", " "));
    break;
  case WATER:
    NC_CHECK(nct_append_attribute(ncid, NC_GLOBAL, "title", "liquid drops", " "));
    break;
  default:
    NC_CHECK(nct_append_attribute(ncid, NC_GLOBAL, "title", medium_str, " "));
  };

  switch (engine) {
  case BHMIE:
    NC_CHECK(nct_append_attribute(ncid, NC_GLOBAL, "title",
				  "computed using Mie theory", " "));
    fprintf(stdout, "Method: Mie theory\n");
    break;
  case AREA_SCALED_BHMIE:
    NC_CHECK(nct_append_attribute(ncid, NC_GLOBAL, "title",
				  "computed using area-scaled Mie theory", " "));
    fprintf(stdout, "Method: area-scaled Mie theory\n");
    break;
  case TMATRIX:
    NC_CHECK(nct_append_attribute(ncid, NC_GLOBAL, "title",
	  "computed using the T-matrix method for horizontally oriented oblate spheroids", " "));
    fprintf(stdout, "Method: T-matrix assuming horizontally oriented spheroids\n");
    break;
  case SELF_SIMILAR_RAYLEIGH_GANS_ORIG:
    NC_CHECK(nct_append_attribute(ncid, NC_GLOBAL, "title",
	  "computed using the Self-Similar Rayleigh Gans Approximation (original method of Hogan and Westbrook 2014)", " "));
    fprintf(stdout, "Method: Self-Similar Rayleigh Gans (Hogan & Westbrook 2014):\n  axial_ratio=%g, kappa=%g, beta=%g, gamma=%g\n",
	    axial_ratio, ssrga_kappa, ssrga_beta, ssrga_gamma);
    break;
  case SELF_SIMILAR_RAYLEIGH_GANS:
    NC_CHECK(nct_append_attribute(ncid, NC_GLOBAL, "title",
	  "computed using the Self-Similar Rayleigh Gans Approximation (Hogan et al. 2017)", " "));
    fprintf(stdout, "Method: Self-Similar Rayleigh Gans (Hogan et al 2017):\n  axial_ratio=%g, alpha_eff=%g, kappa=%g, beta=%g, gamma=%g, zeta1=%g, axial_ratio_monomer=%g\n",
	    axial_ratio, ssrga_alpha_eff, ssrga_kappa, ssrga_beta, ssrga_gamma, ssrga_zeta1,
	    axial_ratio_monomer);
    break;
  case AREA_LOOK_UP_TABLE:
    NC_CHECK(nct_append_attribute(ncid, NC_GLOBAL, "title",
				  "extracted from", " "));
    NC_CHECK(nct_append_attribute(ncid, NC_GLOBAL, "title",
				  lut_filename, " "));  
  }

  NC_CHECK(nct_add_history(ncid, "Computed particle scattering properties", NULL));
  NC_CHECK(nct_add_command_line(ncid, argc, argv));
  NC_CHECK(nct_add_string_attribute(ncid, NC_GLOBAL, "config",
				    rc_sprint(config)));

  if (compute_phase_function) {
    if (n_legendre_coefficients > 0) {
      fprintf(stdout, "Writing phase function and %d Legendre coefficients\n",
	      n_legendre_coefficients);
    }
    else {
      fprintf(stdout, "Writing phase function\n");
    }
  }


  /* WRITE DATA TO NETCDF FILE */

  NC_CHECK(nc_enddef(ncid));

  /* Populate vectors */
  NC_CHECK(NC_PUT_VAR_REAL(ncid, wavelength_id, wavelength));

  NC_CHECK(NC_PUT_VAR_REAL(ncid, temperature_id, temperature));
  if (n_fraction > 0) {
    NC_CHECK(NC_PUT_VAR_REAL(ncid, fraction_id, fraction));
  }
  else {
    n_fraction = 1;
  }
  
  /* Write size coordinate-like variables */
  NC_CHECK(NC_PUT_VAR_REAL(ncid, size_id, size));
  NC_CHECK(NC_PUT_VAR_REAL(ncid, dmax_id, Dmax));
  NC_CHECK(NC_PUT_VAR_REAL(ncid, area_id, area));

  for (j = 0; j < n_fraction; j++) {
    real density = max_density*fraction[j];
    index[0] = j;
    for (i = 0; i < n_size; i++) {
      real mass = (1.0/6.0)*pi*size[i]*size[i]*size[i]*density;
      index[1] = i;
      NC_CHECK(NC_PUT_VAR1_REAL(ncid, mass_id, index+(n_fraction<2), &mass));
    }
  }

  if (calc_fall_speed) {
    if (medium == WATER) {
      for (i = 0; i < n_size; i++) {
	real diameter = size[i];	/* Size is diameter here */
	if (diameter > 0.01) {
	  diameter = 0.01; /* Cap diameter to 1 cm */
	}
	real fall_speed
	  = beard_droplet_fall_speed(diameter,
				     TEMPERATURE_DROP_FALLSPEED_REF,
				     AIR_DENSITY_DROP_FALLSPEED_REF);
	index[0] = i;
	NC_CHECK(NC_PUT_VAR1_REAL(ncid, fall_speed_id, index, &fall_speed));
      }
    }
    else if (medium == ICE) {
      for (j = 0; j < n_fraction; j++) {
	real density = max_density*fraction[j];
	index[0] = j;
	for (i = 0; i < n_size; i++) {
	  real diameter = size[i];	/* Size is diameter here */
	  real mass = (1.0/6.0)*pi*diameter*diameter*diameter*density;
	  /*  Mitchell (1996) relation for polycrystals & aggregates: */
	  /* real area = 0.131*pow(diameter,1.88); */
	  /* Now allow user modification */
	  real area_ratio = area[i]/(0.25*pi*diameter*diameter);
	  if (area_ratio > 1.0) {
	    area_ratio = 1.0;
	  }
	  else if (area_ratio < 0.05) {
	    area_ratio = 0.05;
	  }
	  real fall_speed = heymsfield_westbrook_ice_fall_speed
	    (AIR_DENSITY_ICE_FALLSPEED_REF, mass, diameter, area_ratio);
	  index[1] = i;
	  NC_CHECK(NC_PUT_VAR1_REAL(ncid, fall_speed_id, 
				    index+(n_fraction<2), &fall_speed));
	}
      } 
    }
  }

  if (n_phase_function_angles > 0) {
    int iang;
    for (iang = 0; iang < n_phase_function_angles; iang++) {
      real angle = iang*180.0/(n_phase_function_angles-1.0);
      index[0] = iang;
      NC_CHECK(NC_PUT_VAR1_REAL(ncid, phase_function_angle_id, index, &angle));
    }
  }

  index[0] = 0;
  index[1] = 0;
  index[2] = 0;
  index[3] = 0;
  index[4] = 0;
  index[5] = 0;

  n_total = n_wavelength * n_temperature * n_fraction * n_size;

  if (show_dots) {
    fprintf(stdout, "0%%");
  }

  for (l = 0; l < n_wavelength; l++) {
    if (n_wavelength > 1) {
      index[i_wavelength] = l;
    }
    for (k = 0; k < n_temperature; k++) {
      float refind_r_max, refind_i_max;
      if (n_temperature > 1) {
	index[i_temperature] = k;
      }
      if (medium == WATER) {
	refind_water(wavelength[l], temperature[k],
		     &refind_r_max, &refind_i_max);
      }
      else if (medium == ICE) {
	refind_ice(wavelength[l], temperature[k],
		   &refind_r_max, &refind_i_max);
      }
      else {
	if (n_refind_r == 1) {
	  refind_r_max = refind_r[0];
	  refind_i_max = refind_i[0];
	}
	else {
	  refind_r_max = refind_r[k];
	  refind_i_max = refind_i[k];
	}
      }

      for (j = 0; j < n_fraction; j++) {
	float refind_r, refind_i;
	if (n_fraction > 1) {
	  index[i_fraction] = j;
	}

	if (fraction[j] != 1.0) {
	  maxwell_garnett(fraction[j], refind_r_max, refind_i_max,
			  &refind_r, &refind_i);
	}
	else {
	  refind_r = refind_r_max;
	  refind_i = refind_i_max;
	}
	NC_CHECK(nc_put_var1_float(ncid, refind_r_id, index, &refind_r));
	NC_CHECK(nc_put_var1_float(ncid, refind_i_id, index, &refind_i));

	for (i = 0; i < n_size; i++) {
	  if (n_size > 1) { /* Why is this here? */
	    float cs_ext = 0.0, cs_scat = 0.0, cs_bscat = 0.0, g = 0.0;

	    if (show_dots) {
	      i_particle++;
	      int i_dot = i_particle*80 / n_total;
	      if (i_dot != i_last_dot) {
		if (i_dot == 40) {
		  fprintf(stdout, "50%%");
		}
		else {
		  fprintf(stdout, ".");
		}
		fflush(stdout);
		i_last_dot = i_dot;
	      }
	    }

	    index[i_size] = i;

	    switch (engine) {
	    case BHMIE:
	      {
		float sphere_area = 0.25*pi*size[i]*size[i];
		if (!compute_phase_function) {
		  mie(pi*size[i]/wavelength[l], refind_r, refind_i,
		      &cs_ext, &cs_scat, &cs_bscat, &g);
		}
		else {
		  mie_phase_function(pi*size[i]/wavelength[l], refind_r, refind_i,
				     nang_to_90, &cs_ext, &cs_scat, &cs_bscat, &g,
				     phase_function);
		}
		cs_ext *= sphere_area;
		cs_scat *= sphere_area;
		cs_bscat *= sphere_area;
	      }
	      break;
	    case AREA_SCALED_BHMIE:
	      if (!compute_phase_function) {
		mie(pi*size[i]/wavelength[l], refind_r, refind_i,
		    &cs_ext, &cs_scat, &cs_bscat, &g);
	      }
	      else {
		mie_phase_function(pi*size[i]/wavelength[l], refind_r, refind_i,
				   nang_to_90, &cs_ext, &cs_scat, &cs_bscat, &g,
				   phase_function);
	      }
	      cs_ext *= area[i];
	      cs_scat *= area[i];
	      cs_bscat *= area[i];
	      break;
	    case TMATRIX:
	      tmatrix(size[i], 1.0/axial_ratio, refind_r, refind_i, wavelength[l],
		      &cs_ext, &cs_scat, &cs_bscat, &g);
	      break;
	    case SELF_SIMILAR_RAYLEIGH_GANS_ORIG:
	      {
		/* Use Mie theory for everything then overwrite the
		   backscatter cross-section using the Hogan and
		   Westbrook (2014) formulation */
		float sphere_area = 0.25*pi*size[i]*size[i];
		if (!compute_phase_function) {
		  mie(pi*size[i]/wavelength[l], refind_r, refind_i,
		      &cs_ext, &cs_scat, &cs_bscat, &g);
		}
		else {
		  mie_phase_function(pi*size[i]/wavelength[l], refind_r, refind_i,
				     nang_to_90, &cs_ext, &cs_scat, &cs_bscat, &g,
				     phase_function);
		}
		cs_ext *= sphere_area;
		cs_scat *= sphere_area;
		cs_bscat = self_similar_rayleigh_gans(size[i], axial_ratio,
						      refind_r_max, refind_i_max,
						      fraction[j], wavelength[l],
						      ssrga_kappa,
						      ssrga_beta, ssrga_gamma);
	      }
	      break;
	    case SELF_SIMILAR_RAYLEIGH_GANS:
	      {
		/* Newer Hogan et al. (2017) formulation */

		/* Real and imaginary parts of the dielectric factor */
		real k_r, k_i;
		/* Compute the effective dielectric factor, which is a
		   function of the polarizability of the monomers,
		   which in turn is dependent on their refractive
		   index, their aspect ratio (>1 means
		   prolate/columnar, <1 means oblate/plate-like), and
		   to a lesser extent whether they are hexagonal
		   prisms or spheroids. */
		dielectric_factor_hexagonal(refind_r_max, refind_i_max, axial_ratio_monomer,
					    &k_r, &k_i);
		//dielectric_factor_spheroid(refind_r_max, refind_i_max, axial_ratio_monomer,
		//			   &k_r, &k_i);
		real volume = size[i]*size[i]*size[i]*fraction[j]*(M_PI/6.0);
		real bscat_xs, scat_xs, abs_xs, asym;
		if (!compute_phase_function) {
		  ssrga_cross_sections(wavelength[l], Dmax[i]*ssrga_alpha_eff, volume,
				       k_r, k_i, 
				       ssrga_kappa, ssrga_beta,
				       ssrga_gamma, ssrga_zeta1,
				       &bscat_xs, &scat_xs, &abs_xs, &asym);
		}
		else {
		  real pp[nang];
		  ssrga_phase_function(wavelength[l], Dmax[i]*ssrga_alpha_eff, volume,
				       k_r, k_i, 
				       ssrga_kappa, ssrga_beta,
				       ssrga_gamma, ssrga_zeta1,
				       &bscat_xs, &scat_xs, &abs_xs, &asym,
				       nang, pp);
		  int iang;
		  for (iang = 0; iang < nang; ++iang) {
		    phase_function[iang] = pp[iang];
		  }
		}
		cs_scat  = scat_xs;
		cs_ext   = abs_xs + scat_xs;
		g        = asym;

		/* Overwrite backscatter for nadir/zenith incidence,
		   scaling to convert from radar backscatter to
		   backscatter cross-section */
		cs_bscat = (0.25/M_PI) * ssrga_backscatter(wavelength[l], Dmax[i]*axial_ratio, volume,
							   k_r, k_i, 
							   ssrga_kappa, ssrga_beta,
							   ssrga_gamma, ssrga_zeta1);
	      }
	      break;
	    case AREA_LOOK_UP_TABLE:
	      if (compute_phase_function) {
		fprintf(stderr, "Error: cannot yet compute phase function from look-up table\n");
		return 1;
	      }
	      look_up(&lut, wavelength[l], area[i], &cs_ext, &cs_scat, &cs_bscat, &g);
	      break;
	    default:
	      fprintf(stderr, "Error: scattering engine with code %d not supported\n",
		      engine);
	      return 1;
	    }
	    NC_CHECK(nc_put_var1_float(ncid, ext_id, index, &cs_ext));
	    NC_CHECK(nc_put_var1_float(ncid, scat_id, index, &cs_scat));
	    NC_CHECK(nc_put_var1_float(ncid, bscat_id, index, &cs_bscat));
	    NC_CHECK(nc_put_var1_float(ncid, g_id, index, &g));

	    if (n_phase_function_angles > 0) {
	      size_t count[6] = {1,1,1,1,1,1};
	      count[ndims] = n_phase_function_angles;
	      NC_CHECK(nc_put_vara_float(ncid, phase_function_id, index, count,
					 phase_function));
	    }

	    if (n_legendre_coefficients > 0) {
	      float legendre_coefficients[n_legendre_coefficients];
	      legendre_decomposition(n_phase_function_angles, n_legendre_coefficients,
				     phase_function, legendre_coefficients);
	      /* The asymmetry factor is likely to be more accurately
		 computed than the second term in the Legendre
		 series, so we use g*3.0 here */
	      legendre_coefficients[1] = g*3.0;
	      size_t count[6] = {1,1,1,1,1,1};
	      count[ndims] = n_legendre_coefficients;
	      NC_CHECK(nc_put_vara_float(ncid, legendre_coefficient_id, index, count,
					 legendre_coefficients));
	    }
	  }
	} /* Loop over size */
      } /* Loop over fraction */
    } /* Loop over temperature */
  } /* Loop over wavelength */
  if (show_dots) {
    fprintf(stdout, "100%%\n");
  }

  NC_CHECK(nc_close(ncid));
  return 0;
}
