/* ssrga.c -- Self-Similar Rayleigh-Gans Approximation calculations
 
Author: Robin Hogan <r.j.hogan@ecmwf.int>
Copyright (C) 2016-2019  European Centre for Medium Range Weather Forecasts 
Language: C99

The Self-Similar Rayleigh Gans Approximation computes scattering
properties for ice aggregates; See Hogan et al. (QJRMS 2017) for full
details of the method. Normal usage would be to first call one of the
dielectric_factor_* functions to compute the real and imaginary parts
of the effective dielectric factor "K" of the monomers from which the
aggregate is composed. The aspect ratio of the monomers is provided at
this point. Then pass "K" into one of the ssrga_* functions to compute
the backscatter cross-section, scattering properties or the full phase
function.

 */

#include <math.h>
#include <complex.h>
#include <stdio.h>

#include "pi.h"
#include "ssrga.h"

// Fast version of "pow": slightly less accurate but if we use double
// precision here and store in single precision then there is
// typically no difference
#define POW(a,b) (exp((b)*log(a)))
//#define POW(a,b) pow((a),(b))

/* Factor outside the braces in Eq. 12 of Hogan and Westbrook (2014) */
static inline
real
ssrga_prefactor(real wavenumber, /* Wavenumber (m-1) */
		real k_r,        /* Real part of dielectric factor */
		real k_i,        /* Imag part of dielectric factor */
		real volume)     /* Volume of ice (m3) */
{
  static const real nine_pi_over_sixteen = 9.0 * M_PI / 16.0;

  /* Dielectric factor |K^2| */
  real k2 = k_r*k_r + k_i*k_i;

  real wavenumber2 = wavenumber * wavenumber;
  return nine_pi_over_sixteen * wavenumber2 * wavenumber2 * k2 * volume * volume;
}

/* Compute the first term in the braces in Eq. 4 of Hogan et
   al. (2017), which describes the scattering by the mean ice
   distribution in the particles. FIX - x ~ relevant multiple of PI */
static inline
real
ssrga_envelope_term(real x,     /* Electrical size */
		    real kappa) /* Kurtosis parameter */
{
  real term;
  if (x > 0.0) {
    term = cos(x) * ( (1.0+kappa/3.0)*(1.0/(2.0*x+M_PI) - 1.0/(2.0*x-M_PI))
			   -kappa*(1.0/(2.0*x+3.0*M_PI) - 1.0/(2.0*x-3.0*M_PI)) );
    term *= term;
  }
  else {
    term = 4.0 / (M_PI*M_PI);
  }
  return term;
}

/* Compute the second term in the braces in Eq. 4 of Hogan et
   al. (2017), which describes the scattering by the substructure of
   the ice distribution that differs from one particle to another. FIX
   if x ~ n*PI */
static
real
ssrga_summation_term(real x,     /* Electrical size */
		     real beta,  /* Power-law prefactor */
		     real gamma, /* Power-law exponent (+ve) */
		     real zeta1) /* Wavenumber-1 power scaling */
{
  real sum;

  /* First term is multiplied by zeta1 */
  {
    real term1 = 0.5 / (x + M_PI);
    real term2 = 0.5 / (x - M_PI);
    sum = zeta1 * POW(2.0, -gamma) * (term1*term1 + term2*term2);
  }

  int  jmax = floor(5.0 * x / M_PI + 1.0);
  int  j;
  for (j = 2; j <= jmax; ++j) {
    real term1 = 0.5 / (x + j*M_PI);
    real term2 = 0.5 / (x - j*M_PI);
    sum += POW(2.0*j, -gamma) * (term1*term1 + term2*term2);
  }
  real sin_x = sin(x);
  return beta * sin_x * sin_x * sum;
}


/* Compute the real and imagnary parts of the dielectric factor for a
   sphere using the Clausius-Mossotti formula. */
void
dielectric_factor_sphere(real refind_r, /* Real part refractive index */
			 real refind_i, /* Imag part refractive index */
			 real* k_r,     /* Real part dielectric factor */
			 real* k_i)     /* Imag part dielectric factor */
{
  real_complex refind = refind_r + refind_i * I; 
  real_complex epsilon = refind*refind;
  real_complex k = (epsilon-1.0) / (epsilon+2.0);
  *k_r = creal(k);
  *k_i = cimag(k);
}


/* Compute the real and imaginary parts of the effective dielectric
   factor for a randomly oriented spheroidal monomers using Gans
   theory, where aspect<1 indicates oblate spheroids and aspect>1
   indicates prolate spheroids. */
void
dielectric_factor_spheroid(real refind_r, /* Real part refractive index */
			   real refind_i, /* Imag part refractive index */
			   real aspect,   /* Aspect ratio */
			   real* k_r,     /* Real part dielectric factor */
			   real* k_i)     /* Imag part dielectric factor */
{
  static const real one_third = 1.0 / 3.0;
  static const real two_thirds = 2.0 / 3.0;

  real_complex refind  = refind_r + refind_i * I; 
  real_complex epsilon = refind*refind;

  real l_z;

  if (aspect <= 1.0) {
    /* Oblate spheroid */
    real ecc = sqrt(1.0 - aspect*aspect); /* Eccentricity */
    l_z = (1.0-aspect*asin(ecc)/ecc)/(ecc*ecc);
  }
  else {
    /* Prolate spheroid */
    real ecc = sqrt(1.0 - 1.0/(aspect*aspect));
    l_z = (1.0/(ecc*ecc) - 1.0) * ((1.0 / (2.0*ecc))*log((1.0+ecc)/(1.0-ecc)) - 1.0);
  }
    
  real l_x = 0.5*(1.0 - l_z);
  /* Polarizability along the principal axes, removing the volume
     term; implicitly k_y=k_x */
  real_complex k_z = (epsilon-1.0) / (3.0*l_z*(epsilon-1.0) + 3.0);
  real_complex k_x = (epsilon-1.0) / (3.0*l_x*(epsilon-1.0) + 3.0);
  real_complex k = csqrt(two_thirds * k_x * k_x + one_third * k_z * k_z);
  *k_r = creal(k);
  *k_i = cimag(k);
}


/* Compute the real and imaginary parts of the effective dielectric
   factor for a randomly oriented hexagonal monomers using Westbrook
   (2012) theory, where aspect<1 indicates a hexagonal plate and
   aspect>1 indicates a hexagonal column. */
void
dielectric_factor_hexagonal(real refind_r,/* Real part refractive index */
			    real refind_i,/* Imag part refractive index */
			    real aspect,  /* Aspect ratio */
			    real* k_r,    /* Real part dielectric factor */
			    real* k_i)    /* Imag part dielectric factor */
{
  static const real one_third = 1.0 / 3.0;
  static const real two_thirds = 2.0 / 3.0;
  real_complex refind = refind_r + refind_i * I;
  real_complex epsilon = refind*refind;
  real aspect_power = 0.5 * POW(aspect, -0.9);
  /* Westbrook (QJ 2014) equations 12 and 13 */
  real l_z = 0.5 * ((1.0-3.0*aspect)/(1.0+3.0*aspect) + 1.0);
  real l_x = 0.25* ((1.0-aspect_power)/(1.0+aspect_power) + 1.0);
  /* Polarizability along the principal axes, removing the volume
     term; implicitly k_y=k_x */
  real_complex k_z = (epsilon-1.0) / (3.0*l_z*(epsilon-1.0) + 3.0);
  real_complex k_x = (epsilon-1.0) / (3.0*l_x*(epsilon-1.0) + 3.0);
  real_complex k = csqrt(two_thirds * k_x * k_x + one_third * k_z * k_z);
  *k_r = creal(k);
  *k_i = cimag(k);
}

/* Return the radar backscatter cross-section for the parameters
   provided, with units of "m2". To obtain backscatter per unit solid
   angle (in m2 sr-1), divide the result by 4pi. The meanings of
   kappa, beta, gamma and zeta1 are given in Hogan et al. (2017). The
   variable "d_eff" is equal to the maximum dimension multiplied by
   the effective aspect ratio (denoted alpha_eff by Hogan et
   al. 2017). */
real
ssrga_backscatter(real wavelength, /* Wavelength (m) */
		  real d_eff,      /* Particle dimension in propagation direction (m) */
		  real volume,     /* Volume of material */
		  real k_r,        /* Real part of dielectric param */
		  real k_i,        /* Imag part of dielectric param */
		  real kappa,      /* Kurtosis parameter */
		  real beta,       /* Power law prefactor */
		  real gamma,      /* Power law exponent (+ve) */
		  real zeta1)      /* Wavenumber-1 power scaling */
{
  /* Wavenumber (m-1) */
  real wavenumber = 2.0 * M_PI / wavelength;

  /* Electrical size */
  real x = wavenumber * d_eff;

  return ssrga_prefactor(wavenumber, k_r, k_i, volume)
    * (  ssrga_envelope_term(x, kappa) 
       + ssrga_summation_term(x, beta, gamma, zeta1));
}

/* Return the radar backscatter cross-section, the scattering and
   absorption cross-sections, the asymmetry factor and the phase
   function, for the parameters provided. The phase function is
   normalized such that the integral from 0 to pi of the phase
   function multiplied by the sine of the scattering angle is
   one. Except for backscatter, the scattering properties are computed
   assuming random particle orientation with respect to the direction
   of propagation of the incident radiation. If you require the
   backscatter for nadir or zenith incidence then you may wish to call
   ssrga_backscatter separately for that one variable. */
void
ssrga_phase_function(real wavelength, /* Wavelength (m) */
		     real d_eff,      /* Particle dimension in propagation direction (m) */
		     real volume,     /* Volume of material */
		     real k_r,        /* Real part of dielectric_param */
		     real k_i,        /* Imag part of dielectric_param */
		     real kappa,      /* Kurtosis parameter */
		     real beta,       /* Power law prefactor */
		     real gamma,      /* Power law exponent (+ve) */
		     real zeta1,      /* Wavenumber-1 power scaling */
		     real* bscat_xs,  /* Backscatter cross section */
		     real* scat_xs,   /* Scattering cross section */
		     real* abs_xs,    /* Absorption cross section */
		     real* asym,      /* Asymmetry factor */
		     int nang,        /* Number of scattering angles */
		     real pp[nang])   /* Scattering phase function */
{
  /* Wavenumber (m-1) */
  real wavenumber = 2.0 * M_PI / wavelength;

  /* Electrical size */
  real x = wavenumber * d_eff;

  *abs_xs = 3.0 * volume * wavenumber * fabs(k_i);

  /* Angular interval, in radians */
  real dang = M_PI / (nang - 1.0);

  real nang_recip = M_PI / (nang - 1.0);

  int iang;

  real sum   = 0.0;
  real sum_g = 0.0;

  /* Improved precision by working back from largest scattering angle */
#pragma omp parallel for reduction(+: sum,sum_g)
  for (iang = nang-1; iang >= 0; --iang) {
    real ang = iang * nang_recip;
    real x_eff = x*sin(0.5*ang);
    pp[iang] = ssrga_envelope_term(x_eff, kappa)
      + ssrga_summation_term(x_eff, beta, gamma, zeta1);
    real cos_ang = cos(ang);
    pp[iang] *= (1.0+cos_ang*cos_ang)*0.5;
    if (iang > 0 && iang < nang-1) {
      real pp_sin_ang_dang = pp[iang]*sin(ang)*dang;
      sum   += pp_sin_ang_dang;
      sum_g += pp_sin_ang_dang * cos_ang;
    }
  }

  *asym = sum_g / sum;

  real prefactor = ssrga_prefactor(wavenumber, k_r, k_i, volume);
  *bscat_xs = prefactor * pp[nang-1];
  *scat_xs  = 0.5 * prefactor * sum;

  real pp_scaling = 1.0 / sum;
  /* Renormalize phase function*/
  for (iang = 0; iang < nang; ++iang) {
    pp[iang] *= pp_scaling;
  }
}

/* Return the radar backscatter cross-section, the scattering and
   absorption cross-sections, and the asymmetry factor, for the
   parameters provided. Except for backscatter, the scattering
   properties are computed assuming random particle orientation with
   respect to the direction of propagation of the incident
   radiation. If you require the backscatter for nadir or zenith
   incidence then you may wish to call ssrga_backscatter separately
   for that one variable. */
void
ssrga_cross_sections(real wavelength, /* Wavelength (m) */
		     real d_eff,      /* Particle dimension in propagation direction (m) */
		     real volume,     /* Volume of material */
		     real k_r,        /* Real part of dielectric_param */
		     real k_i,        /* Imag part of dielectric_param */
		     real kappa,      /* Kurtosis parameter */
		     real beta,       /* Power law prefactor */
		     real gamma,      /* Power law exponent (+ve) */
		     real zeta1,      /* Wavenumber-1 power scaling */
		     real* bscat_xs,  /* Backscatter cross section */
		     real* scat_xs,   /* Scattering cross section */
		     real* abs_xs,    /* Absorption cross section */
		     real* asym)      /* Asymmetry factor */
{
  /* Work out the number of angles needed to compute the phase
     function with sufficient detail to then compute the scattering
     cross section and asymmetry factor from it */
  int nang = (int)(50.0*d_eff/wavelength);
  nang = nang < 19 ? 19 : nang;
  real pp[nang];
  ssrga_phase_function(wavelength, d_eff, volume, k_r, k_i,
		      kappa, beta, gamma, zeta1,
		      bscat_xs, scat_xs, abs_xs, asym,
		      nang, pp);
}
